//=============================================================================
// AssignExprsTree.g - Based on AssignExprs, build a Tree for 
//                     Assignment Statements and Arithmetic Expressions
//                     Including Function Invocation and 1-D Arrays
//                     Then, use a tree-walker to walk the AST and emit
//                       code for an imaginary stack machine...
//                     Semantics assume a stack machine with tagged entries
//                       Tags:  value, address, fcn-address
//
// Author:  Bary W Pollack
// Date:    Mar. 20, 2004 - 17:00
//
//=============================================================================

// Import the necessary classes
{
import java.io.*;
import java.util.*;
import antlr.debug.misc.ASTFrame;
}


//-----------------------------------------------------------------------------
// Define a Parser, call it AssignExprsTreeParser
//-----------------------------------------------------------------------------

class AssignExprsTreeParser extends Parser;

options {
    defaultErrorHandler = false;    // Don't generate parser error handlers
    buildAST = true;                // Turn on AST tree construction
}

tokens {
    ARRAY; CALL; PROGRAM; 
}


// Define some methods and variables to use in the generated parser
{
    // Define a main
    public static void main(String[] args) {
        // Use a try/catch block for parser exceptions
        try {
            // if we have at least one command-line argument
            if (args.length > 0 ) {
                System.out.println();
                System.out.println("AssignExprsTree Parsing...");
                System.out.println();

                // For each directory/file specified on the command line
                for (int i = 0; i < args.length; i++)
                    doFile(new File(args[i]));  // parse it
            } else
                System.out.println("Usage: java AssignExprsTreeParser [ file | <directory name> ]...");

        } catch (Exception e) {
            System.out.println("Exception: " + e);
            e.printStackTrace(System.out);   // so we can get a stack trace
        }
    }


    // This method decides what action to take based on the
    // type of file we are looking at -- directory or source file
    public static void doFile(File f) throws Exception {
        // If this is a directory, walk each file/dir in that directory
        if (f.isDirectory()) {
            String files[] = f.list();
            for (int i = 0; i < files.length; i++)
                doFile(new File(f, files[i]));        
        // otherwise, if this is a source file, parse it
        } else if ((f.getName().length() >= 5) &&
                   f.getName().substring(f.getName().length() - 4).equals(".txt")) {
            ///System.out.println("  File: " + f.getAbsolutePath());  //FOO
            System.out.println("  File: " + f.getName());
            parseFile(new FileInputStream(f));
        }
    }

    // Here's where we do the real work...
    public static void parseFile(InputStream s) throws Exception {
        System.out.println();
        try {
            // Create a scanner that reads from the input stream passed to us
            AssignExprsLexer lexer = new AssignExprsLexer(s);

            // Create a parser that reads from the scanner
            AssignExprsTreeParser parser = new AssignExprsTreeParser(lexer);

            // Start parsing at the program rule
            pSymbols = new TreeMap();  // initialize the symbol table
            parser.program();
            System.out.println();
            System.out.println("* Parsing Successful *");


            // Create an AST for the input expression
            AST tree = parser.getAST();
            System.out.println();
            System.out.println("Tree: " + tree.toStringTree());
            System.out.println();
            System.out.println("* Tree Construction Successful *");
            System.out.println();

            // Create an ASTFrame to show the tree structure in 2D
            if (true) {
                ASTFrame frame = new ASTFrame("Tree for: " + tree.toStringList(), tree);
                frame.setVisible(true);
            }

            // Walk the AST to create the semantics (evaluate the expression)
            AssignExprsTreeWalker treeWalker = new AssignExprsTreeWalker();
            treeWalker.program(tree);

            System.out.println();
            System.out.println("* Tree-Walk Successful *");
            System.out.println();

        } catch (Exception e) {
            System.out.println("Parser exception: " + e);
            System.out.println();
            e.printStackTrace();   // so we can get a stack trace

        } finally {
            if (false) {           // control display of symbol table  //FOO
                System.out.println();
                displaySymbols();
                System.out.println();
            }
        }
    }

    // A tiny symbol table utility
    // Contains symbol, its type (kind), source line number where last used

    public static final char IDENTIFIER = 'I';
    public static final char INTEGER    = 'N';

    public class Symbol {
        Symbol(String sName, char cKind, int nLno) {
            this.sName = sName;  this.cKind = cKind;  this.nLno = nLno;
        }
        public String toString() {
            return "<" + sName + " " + cKind + " " + nLno + ">";
        }
        private String sName;
        private char   cKind;
        private int    nLno;
    }

    public static TreeMap pSymbols;  // this is the symbol table

    static void displaySymbols() {
        System.out.println("Symbol Table (" + pSymbols.size() + " entries)");
        int n = 1;
        Set set = pSymbols.entrySet();
        for (Iterator i = set.iterator(); i.hasNext(); n++) {
            Map.Entry e = (Map.Entry) i.next();
            String s = ((String)e.getKey()) + "        ";
            System.out.print("   " + (n + "    ").substring(0,3));
            System.out.print("  " + s.substring(0, 8) + " ");
            System.out.println((Symbol)e.getValue());
        }
    }

    static void sop(String s) {         //FOO
        if (true)                       //FOO
        System.out.println(s); 
    }

}

// ==================================================================

program
  : programBody 
      { sop("program");
        #program = #([PROGRAM, "PROGRAM"], program); 
      }
    // end-of-file
  ;

programBody
  : "begin"!
       statementList
    "end"!
  ;

statement
  : (IDENT (LPAREN|SEMI))=> procedureCallStatement
  | assignmentStatement
  ;

statementList
  : statement statementList
  | // nothing
  ;

assignmentStatement
  { String varx = "who"; }
  : varx=variableReference EQUALS^ expression SEMI!
      { sop("assignmentStatement: " + varx); sop(""); }
  ;

procedureCallStatement
  : ident:IDENT actualParameters SEMI!
      { System.out.println("ident=  " + ident.getText());    //FOO
        pSymbols.put(ident.getText(), 
                     new Symbol(ident.getText(), IDENTIFIER, ident.getLine()));
        sop("procedureCallStatement: " + ident.getText()); sop("");
        #procedureCallStatement = #([CALL, "CALL"], procedureCallStatement);
      }
  ;

actualParameters
  : LPAREN! (expression (COMMA! expression)*)? RPAREN!
  ;

procedureCall
  : ident:IDENT actualParameters
      { System.out.println("ident=  " + ident.getText());    //FOO
        pSymbols.put(ident.getText(), 
                     new Symbol(ident.getText(), IDENTIFIER, ident.getLine()));
        sop("procedureCall: " + ident.getText());
        #procedureCall = #([CALL, "CALL"], procedureCall);
      }
  ;

expression
  : multiplyingExpression ((PLUS^|MINUS^) multiplyingExpression)*
  ;

multiplyingExpression
  : signedExpression ((TIMES^|DIV^) signedExpression)*
  ;

signedExpression
  : (PLUS^|MINUS^)* primitiveElement
  ;

primitiveElement
  { String varx = ""; }
  : (IDENT LPAREN)=> procedureCall
  | varx=variableReference
  | constantValue
  | LPAREN! expression RPAREN!
  ;

variableReference returns [ String varx ]
  { varx = ""; }
  : ident:IDENT
      { System.out.println("ident=  " + ident.getText());    //FOO
        pSymbols.put(ident.getText(), 
                     new Symbol(ident.getText(), IDENTIFIER, ident.getLine()));
        varx = ident.getText();
      }
    ( LBRACKET! expression RBRACKET!
        { #variableReference = #([ARRAY, "ARRAY"], variableReference); }
    )?
  ;

constantValue
  : intlit:INTLIT
      { System.out.println("intlit= " + intlit.getText());    //FOO
        pSymbols.put(intlit.getText(), 
                     new Symbol(intlit.getText(), INTEGER, intlit.getLine()));
      }
  ;


//----------------------------------------------------------------------------
// The AssignExprsTree scanner
//----------------------------------------------------------------------------

class AssignExprsLexer extends Lexer;

options {
    charVocabulary = '\0'..'\377';
    testLiterals = false;        // don't automatically test for literals
    k = 2;                       // two characters of lookahead
}


// Single-line comments
COMMENT
  : "//" (~('\n'|'\r'))*
    { $setType(Token.SKIP); }
  ;


// Literals
protected DIGIT : '0'..'9';

INTLIT : (DIGIT)+ ;


// Keywords are literals in the parser grammar

// Operators
SEMI     : ';' ;
COMMA    : ',' ;
EQUALS   : '=' ;
LBRACKET : '[' ;
RBRACKET : ']' ;
LPAREN   : '(' ;
RPAREN   : ')' ;
PLUS     : '+' ;
MINUS    : '-' ;
TIMES    : '*' ;
DIV      : '/' ;


// Whitespace -- ignored
WS
  : ( ' '
    | '\t'
    | '\f'

    // Handle newlines
    | ( "\r\n"  // Evil DOS
      | '\r'    // Macintosh
      | '\n'    // Unix (the right way)
      )
      { newline(); }
    )
    { $setType(Token.SKIP); }
  ;

// Multiple-line comments
ML_COMMENT
  : "/*"
    (    /* '\r' '\n' can be matched in one alternative or by matching
           '\r' in one iteration and '\n' in another.  I am trying to
            handle any flavor of newline that comes in, but the language
            that allows both "\r\n" and "\r" and "\n" to all be valid
            newline is ambiguous.  Consequently, the resulting grammar
            must be ambiguous.  I'm shutting this warning off.
        */
        options {
          generateAmbigWarnings = false;
        }
      :
        { LA(2)!='/' }? '*'
      |  '\r' '\n'       { newline(); }
      |  '\r'            { newline(); }
      |  '\n'            { newline(); }
      |  ~('*'|'\n'|'\r')
    )*
    "*/"
    { $setType(Token.SKIP); }
  ;


// An identifier.  Note that testLiterals is set to true.  This means
// that after we match the rule, we look in the literals table to see
// if it's a literal or really an identifer
IDENT
  options { testLiterals=true; }
  : ('a'..'z'|'A'..'Z') ('a'..'z'|'A'..'Z'|'0'..'9')*
  ;


//----------------------------------------------------------------------------
// The AssignExprsTree tree-walker
//----------------------------------------------------------------------------

class AssignExprsTreeWalker extends TreeParser;

options {
    importVocab = AssignExprsTreeParser;
}

/*
  These rule definitions all use the AST syntax:
    #(ROOT child.1 child.2 ... child.n);
*/

program
  { AST ast = null;}
  : #(PROGRAM (ast=expression)*)
      { E.emit("END"); }
  ;

expression returns [ AST r ]
  { AST a, b = null; r = _t; 
    /* ..... debugging - to see what's in r .....
    System.out.print("expression. r = " + r); 
    if (r != null)
       System.out.println("\t   tSL: " + r.toStringList());
    else
       System.out.println();
    ..... end debugging ..... */
  }

  : #(CALL id01:IDENT (a=expression)*
      { E.emit("LOADF " + id01.getText());
        E.emit("CALL\t\t# call " + id01.getText());
        System.out.println(); 
      }
    )

  | #(EQUALS a=expression b=expression)
      { E.emit("STORE\t\t# " + a + " <- " + b); 
        System.out.println(); 
      }

  | #(PLUS   a=expression b=expression)
      { E.emit("ADD    \t\t# " + a + " + " + b); 
      }

  | #(MINUS  a=expression (b=expression)?)
      { if (b != null)
           E.emit("SUB    \t\t# " + a + " - " + b); 
        else
           E.emit("NEG    \t\t# -" + a); r = a; 
      }

  | #(TIMES  a=expression b=expression)
      { E.emit("MUL    \t\t# " + a + " * " + b);
      }

  | #(DIV    a=expression b=expression)
      { E.emit("DIV    \t\t# " + a + " / " + b); 
      }

  | #(ARRAY  ida:IDENT a=expression)
      { E.emit("LOADA " + ida.getText() + " \t# " + ida.getText());
        E.emit("ARRAY\t\t# " + ida + "[" + a + "]"); r = ida; 
      }

  | id:IDENT
      { r = id; 
        E.emit("LOADA " + id.getText() + " \t# " + id.getText()); 
      }

  | num:INTLIT
      { r = num;
        E.emit("LOADI " + num.getText() + " \t# " + num.getText()); 
      }
  ;

// EOF
